import os
import tempfile
from typing import Literal, TypedDict

from langchain.chat_models import init_chat_model
from langchain_core.messages import AIMessage, SystemMessage, HumanMessage
from langchain_core.runnables import RunnableConfig
from langgraph.graph import END, START, MessagesState, StateGraph
from langgraph.checkpoint.memory import MemorySaver 
from langchain_core.runnables.graph_png import PngDrawer

from python_toolkit import PythonRunnerTool
from sql_toolkit import SQLDatabaseTool, list_tables, call_get_schema, check_query
from utils import show_image

react_system_prompt = """
You are an agent designed to interact with a SQL database.
Given an input question, create a syntactically correct {dialect} query to run,
then look at the results of the query and return the answer. Unless the user
specifies a specific number of examples they wish to obtain, always limit your
query to at most {top_k} results.

You also have available, a python tool ("run_python_tool") that can execute
python code and return the standard output. You can use this for any
computation and transformation of the results. 

You can order the results by a relevant column to return the most interesting
examples in the database. Never query for all the columns from a specific table,
only ask for the relevant columns given the question.

DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.
"""
def react_node(state: MessagesState, config: RunnableConfig):

    # generate a system message.
    cfg = config['configurable'] # pyright: ignore[reportTypedDictNotRequiredAccess]
    sql_toolkit = cfg['sql_toolkit']
    system_message = SystemMessage(react_system_prompt.format(
        dialect=sql_toolkit.dialect,
        top_k=5
    ))
    python_toolkit = cfg['python_runner']

    # We do not force a tool call here, to allow the model to
    # respond naturally when it obtains the solution.
    run_query_tool = sql_toolkit.run_query_tool
    llm_with_tools = sql_toolkit.llm.bind_tools([run_query_tool, python_toolkit.tool()])
    response = llm_with_tools.invoke([system_message] + state["messages"])

    return {"messages": [response]}

def route_tool(state: MessagesState) -> Literal[END, "check_query", "run_python"]:  # pyright: ignore[reportInvalidTypeForm]
    messages = state["messages"]
    last_message = messages[-1]
    assert isinstance(last_message, AIMessage)
    tool_calls = last_message.tool_calls
    if len(tool_calls) == 0:
        return END
    elif len(tool_calls) > 1:
        calls = [ t['name'] for t in tool_calls ]
        raise NotImplementedError(f'Not handling multiple tool calls {calls}')
    elif tool_calls[0]['name'] == 'sql_db_query':
        return 'check_query'
    elif tool_calls[0]['name'] == 'run_python_tool':
        return 'run_python'
    else:
        # should not be here.
        assert False

def should_get_table_meta(_: MessagesState, config: RunnableConfig) -> Literal["react_node", "list_tables"]:
    if config['configurable']['first_time']: # type: ignore
        return 'list_tables'
    else:
        return 'react_node'

class SQLAgentConfigSchema(TypedDict):
    first_time: bool
    python_runner: PythonRunnerTool
    sql_toolkit: SQLDatabaseTool

class SQLAgent:

    def __init__(self, sql_db_uri: str, python_tool: PythonRunnerTool, model:str = "openai:gpt-4.1"):

        # initialization.
        self.llm = init_chat_model(model)
        self.sql_tool = SQLDatabaseTool(db_uri=sql_db_uri, llm=self.llm)
        self.python_tool = python_tool

        # graph building.
        self._build_graph()

        # initialize runtime configuration.
        self._init_config()

    def _build_graph(self):

        builder = StateGraph(MessagesState, config_schema=SQLAgentConfigSchema)
        builder.add_node(list_tables)
        builder.add_node(call_get_schema)
        builder.add_node(self.sql_tool.get_schema_tool_node)
        builder.add_node(react_node)
        builder.add_node(check_query)
        builder.add_node(self.sql_tool.run_query_tool_node)
        builder.add_node(self.python_tool.tool_node())

        builder.add_conditional_edges(START, should_get_table_meta)
        builder.add_edge("list_tables", "call_get_schema")
        builder.add_edge("call_get_schema", "get_schema")
        builder.add_edge("get_schema", "react_node")
        builder.add_conditional_edges(
            "react_node",
            route_tool,
        )
        builder.add_edge("check_query", "run_query")
        builder.add_edge("run_query", "react_node")
        builder.add_edge("run_python", "react_node")

        self.memory = MemorySaver()
        self.agent = builder.compile(checkpointer=self.memory)

    def print_ascii(self):
        self.agent.get_graph().print_ascii()

    def display_graph(self):

        with tempfile.TemporaryDirectory() as tmpdir:
            drawer = PngDrawer()
            png_path = os.path.join(tmpdir, '_lang_graph.png')
            drawer.draw(self.agent.get_graph(), png_path)
            show_image(png_path)

    def _init_config(self):

        self.config = {
                'configurable': {
                    'thread_id': 1,
                    'first_time': True,
                    'python_runner': self.python_tool,
                    'sql_toolkit': self.sql_tool
                }
        }

    def chat(self, question:str, debug=False):

        if debug:
            for step in self.agent.stream(
                {"messages": [HumanMessage(question)]},
                stream_mode="values",
                config=self.config # type: ignore
            ):
                step["messages"][-1].pretty_print()
        else:
            result = self.agent.invoke(
                {"messages": [HumanMessage(question)]},
                config=self.config # type: ignore
            )
            result['messages'][-1].pretty_print()

        # don't run steps needed only the first time.
        self.config['configurable']['first_time'] = False




import sys
import pdb
import subprocess
import os
import logging
import tempfile, tarfile
from typing import Optional, List
from io import BytesIO

import docker
import docker.errors
from docker import DockerClient
from langgraph.prebuilt import ToolNode
from pydantic import BaseModel, Field
from langchain_core.runnables import RunnableConfig
from langchain_core.tools import StructuredTool
from langchain.output_parsers import PydanticOutputParser
from langchain_core.tools import BaseTool, ArgsSchema, BaseToolkit
import yaml

from agentrun_plus import AgentRun, UVInstallPolicy
from loguru import logger

class PythonToolInput(BaseModel):
    python_code: str = Field(description=" Python code to run provided as a string.")
    artifacts_abs_paths: List[str] = Field(description="""
    If the python code is generating a artifact (such as a plot), specify the
    absolute path of the files containing the artifacts.
    """)

class PythonToolOutput(BaseModel):
    stdout: str = Field(description="Text output of the Python Program")
    user_artifacts_abs_paths: List[str] = Field(description="""
    Local paths of artifacts (if any) generated by the Python program, corresponding to the input `artifacts_abs_path` parameter. This is only meant for user display. You can continue to use the original artifact locations for further use in downstream tool calls.
    """)

PYTHON_TOOL_DESCRIPTION="""
Runs the specified python code provided and returns the standard output
text. This tool doesn't have any display capabilities, and hence, if you
wish to generate a plot, store the plot/figure into a PNG file and specify
its absolute path in the `artifacts_abs_paths` parameter. 

Your code will run in a docker environment with a basic python base image
(`python:3.12.2-slim-bullseye`) and the following packages installed:

{packages}

So, restrict your code accordingly.

Always store artifacts such as plots in the `/home/pythonuser/artifacts`
folder. The `artifacts_abs_paths` parameter allows you to generate and pass
multiple artifact paths. Pass an empty list if no artifacts are generated by
the program. 

The tool returns modified locations that you can pass to the user, but you
can continue to refer to the original locations (in `/home/pythonuser/artifacts`) for
referencing in downstream invocations of the tool.

The tool output follows this schema:
{output_schema}
"""
class DockerConfig:

    def __init__(self, 
                 docker_compose_yml:Optional[str]=None,
                 docker_compose_cmd: Optional[List[str]]=None,
    ):

        if not docker_compose_yml:
            docker_compose_yml = os.path.join(
                    os.path.dirname(os.path.abspath(__file__)),
                    "code_runner", 
                    "docker-compose.yml"
                    )
        if not docker_compose_cmd:
            self.docker_compose_cmd = ['docker', 'compose']
        else:
            self.docker_compose_cmd = docker_compose_cmd
        with open(docker_compose_yml, 'rt') as f:
            self.docker_yml = yaml.load(f, Loader=yaml.Loader)
        self.code_runner_path = os.path.dirname(docker_compose_yml)
        self.log = logger.bind(name="Docker")
        self.log.remove()
        self.log.add(
                sys.stderr,
                format="{time} {level} {message}",
                level="INFO",
                enqueue=True
        )
        self.running_on_enter = False

    @staticmethod
    def is_docker_installed():
        try:
            client = docker.from_env()
            _ = client.version()
            return True
        except docker.errors.DockerException:
            return False

    def is_container_running(self):

        client = docker.from_env()

        for _, config in self.docker_yml['services'].items():
            container_name = config['container_name']
            try:
                container = client.containers.get(container_name)
                match container.status:
                    case "running":
                        # if the container is running we are good ...
                        pass
                    case "exited":
                        # if the container was stopped remove it before continuing.
                        self.log.info(f'Stopping container {container_name} ...')
                        container.remove()
                    case _:
                        # otherwise, assume it can started again.
                        return False
            except docker.errors.NotFound:
                self.log.error(f"Container '{container_name}' not found.")
                return False
            except docker.errors.DockerException as e:
                self.log.error(f"Error: {e}")
                return False

        return True

    def __enter__(self):

        if self.is_container_running():
            self.log.info(f'Docker container is already running ...')
            self.running_on_enter = True
            return self
        else:
            self.running_on_enter = False
            self.log.info('Starting Docker Container ...')
            subprocess.run(
                    self.docker_compose_cmd + ['up', '--build', '-d'], 
                    cwd=self.code_runner_path,
                    check=True
                    )
            return self

    def get_container_name(self, service_name: str):
        return self.docker_yml['services'][service_name]['container_name']

    def get_requirements_txt(self):
        reqs_file = os.path.join(self.code_runner_path, 'requirements.txt')
        assert os.path.isfile(reqs_file)
        return reqs_file
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        # don't stop the docker container if it was running on entry.
        if not self.running_on_enter:
            self.log.info('Stopping Python Container ...')
            subprocess.run(
                    self.docker_compose_cmd + ['down'],
                    cwd=self.code_runner_path,
                    check=True
                    )

class DockerContainerNotFound(RuntimeError):
    pass

class PythonRunnerToolContext():

    def __init__(self, 
                 docker_config: DockerConfig,
                 tmpdir: str, 
                 ignore_dependencies: Optional[List[str]]=None,
                 ignore_unsafe_functions: Optional[List[str]]=None,
                 debug=False
    ):
        """
        Initialize the PythonRunnerTool.

        Args:
            docker_config: A DockerConfig instance that starts the docker
                image using a context manager.
            tempdir: A temporary folder to store temporary files (to be deleted
                on program exit). Use the tempfile.TemporaryDirectory context
                manager for this.
        """
        # latch parameters.
        self.debug = debug
        self.client: DockerClient = docker.from_env()
        self.ignore_dependencies = ignore_dependencies
        self.ignore_unsafe_functions = ignore_unsafe_functions
        self.tmpdir = tmpdir

        # initializations.
        self._tool_node: Optional[ToolNode] = None
        self._tool: Optional[StructuredTool] = None
        self.log = logging.getLogger('PYTHON_RUNNER')

        # check if Docker is running by the time this context is created, and
        # intialize AgentRun.
        if not docker_config.is_container_running():
            raise DockerContainerNotFound(f"Docker container service 'python_runner' isn't running!")
        self.docker_config = docker_config
        container_name = self.docker_config.get_container_name("python_runner")
        self.container = self.client.containers.get(container_name)
        self.python_runner = AgentRun(
                container_name=container_name,
                cached_dependencies = ['sqlalchemy'],
                install_policy=UVInstallPolicy(),
                cpu_quota=100000,
                default_timeout=100,
                log_level = 'INFO' if self.debug is True else 'WARNING',
                )


    def execute_code(self, code: str):
        return self.python_runner.execute_code_in_container(
                code, 
                ignore_dependencies=self.ignore_dependencies,
                ignore_unsafe_functions=self.ignore_unsafe_functions
        )

    def configure(self, config: RunnableConfig):
        config['configurable']['python_runner'] = self  # pyright: ignore[reportTypedDictNotRequiredAccess]

    def get_requirements_txt(
            self,
            fmt="markdown"
    ) -> str:
        requirements_txt = self.docker_config.get_requirements_txt()
        lines: List[str] = []
        with open(requirements_txt, 'rt') as f:
            lines = f.readlines()
        match fmt:
            case 'markdown':
                lines = [ f'  - {line.rstrip()}' for line in lines ]
                return '\n'.join(lines)
            case _: 
                raise RuntimeError(f'Unknown format: {format}!')

    def copy_code_to_container(
            self,
            python_code: str,
            dst_path
    ):
        """Copy Python code to the container.
        Args:
            python_code: Python code to copy
            dst_path: Full destination path (incl. file name) inside the container.
        """
        script_path, script_name = os.path.split(dst_path)

        # write the python code to a temporary file.
        temp_script_path = os.path.join(self.tmpdir, script_name)
        with open(temp_script_path, "w") as file:
            file.write(python_code)

        # create a tar stream and add the file to it.
        tar_stream = BytesIO()
        with tarfile.open(fileobj=tar_stream, mode="w") as tar:
            tar.add(temp_script_path, arcname=script_name)
        tar_stream.seek(0)

        # write the file into the container.
        exec_result = self.container.put_archive(path=script_path, data=tar_stream)
        if not exec_result:
            raise RuntimeError(f'Unable to copy code to {dst_path}!')

    def copy_file_to_container(
        self, 
        src_path: str,
        dst_folder: str
    ):
        """Copy Python code to the container.
        Args:
            src_path: Path of the file to copy into container.
            dst_folder: The destination folder inside the container.
        """
        # check if the source file exists.
        if not os.path.isfile(src_path):
            raise RuntimeError(f'{src_path} is not a valid file!')

        # create a tar archive and add the source file.
        tar_stream = BytesIO()
        with tarfile.open(fileobj=tar_stream, mode="w") as tar:
            tar.add(src_path, arcname=os.path.basename(src_path))
        tar_stream.seek(0)

        exec_result = self.container.put_archive(path=dst_folder, data=tar_stream)
        if not exec_result:
            raise RuntimeError(f'Unable to copy {src_path} into container!')

    def copy_file_from_container(
            self, 
            src_path: str,
            dst_folder: Optional[str]=None
            ) -> str:

        # get the archive
        stream, _ = self.container.get_archive(src_path)

        # use temporary folder itself if destination isn't specified.
        if not dst_folder:
            dst_folder = os.path.join(self.tmpdir)

        with tempfile.NamedTemporaryFile(delete=False) as tmpfp:
            try:
                # write the tar stream on to a temporary file ...
                for chunk in stream:
                    tmpfp.write(chunk)
                # close it (the file won't get deleted).
                tmpfp.close()

                # extract it.
                with tarfile.open(tmpfp.name) as tar:
                    tar.extractall(dst_folder)
            finally:
                # delete the tar file eventually.
                os.unlink(tmpfp.name)

        # return the destination file name.
        dst_path = os.path.join(dst_folder, os.path.basename(src_path))
        assert os.path.isfile(dst_path)
        return dst_path

class PythonRunnerTool(BaseTool):

    # overrides from BaseTool
    name: str = "run_python_tool"
    args_schema: ArgsSchema | None = PythonToolInput 
    description: str = ""   # fill this in during construction.
    return_direct: bool = True

    # our configuration.
    config: PythonRunnerToolContext
    
    def _run(
            self,
            python_code: str,
            artifacts_abs_paths: List[str]
    ):
        python_runner = self.config
        output = python_runner.execute_code(python_code)
        local_files = []
        for cont_path in artifacts_abs_paths:
            try:
                # copy the file locally
                local_files.append(python_runner.copy_file_from_container(cont_path))
            except docker.errors.NotFound: 
                raise RuntimeError(f"{cont_path} is inaccessible. The program may have failed - here's the output:\n {output}")
        return PythonToolOutput(
                stdout=output,
                user_artifacts_abs_paths=local_files
        )

class PythonRunnerToolkit(BaseToolkit):
    tools: List[BaseTool]

    @classmethod
    def from_context(
            cls,
            config: PythonRunnerToolContext = Field(exclude=True)
    ):
        output_schema = PydanticOutputParser(pydantic_object=PythonToolOutput).get_output_jsonschema()
        packages = config.get_requirements_txt(fmt='markdown')
        return cls(
                tools=[
                    PythonRunnerTool(
                        description=PYTHON_TOOL_DESCRIPTION.format(
                            output_schema=output_schema,
                            packages=packages,
                        ),
                        config=config
                    )
                ]
            )

    def get_tools(self) -> List[BaseTool]:
        return self.tools



import IPython

from python_toolkit import PythonRunnerTool
from sql_direct_agent import SQLAgent

def main():
    
    with PythonRunnerTool() as python_tool:
        
        agent = SQLAgent('sqlite:///Chinook.db', python_tool)
        IPython.embed()

if __name__ == "__main__":
    main()



import IPython
from langchain_community.utilities import SQLDatabase
from langchain_community.agent_toolkits import SQLDatabaseToolkit
from langchain.chat_models import init_chat_model
from langgraph.prebuilt import create_react_agent

# load database.
db = SQLDatabase.from_uri("sqlite:///Chinook.db")
print(f"Dialect: {db.dialect}")
print(f"Available tables: {db.get_usable_table_names()}")
print(f'Sample output: {db.run("SELECT * FROM Artist LIMIT 5;")}')

# initialize chat model.
llm = init_chat_model("openai:gpt-4.1")

# get and analyze tools.
toolkit = SQLDatabaseToolkit(db=db, llm=llm)
tools = toolkit.get_tools()
for tool in tools:
    print(f"{tool.name}: {tool.description}\n")

# create a react agent and run the model.
system_prompt = """
You are an agent designed to interact with a SQL database.
Given an input question, create a syntactically correct {dialect} query to run,
then look at the results of the query and return the answer. Unless the user
specifies a specific number of examples they wish to obtain, always limit your
query to at most {top_k} results.

You can order the results by a relevant column to return the most interesting
examples in the database. Never query for all the columns from a specific table,
only ask for the relevant columns given the question.

You MUST double check your query before executing it. If you get an error while
executing a query, rewrite the query and try again.

DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the
database.

To start you should ALWAYS look at the tables in the database to see what you
can query. Do NOT skip this step.

Then you should query the schema of the most relevant tables.
""".format(
    dialect=db.dialect,
    top_k=5,
)

agent = create_react_agent(
    llm,
    tools,
    prompt=system_prompt,
)

question = "Which sales agent made the most in sales in 2009?"

for step in agent.stream(
    {"messages": [{"role": "user", "content": question}]},
    stream_mode="values",
):
    step["messages"][-1].pretty_print()

IPython.embed()



import matplotlib.pyplot as plt
import matplotlib.image as mpimg

def show_image(image_path: str):

    img = mpimg.imread(image_path)

    plt.imshow(img)
    plt.axis("off")  # Hide axes for a clean display
    plt.show()




import os, sys
import urllib.request
import tempfile

import IPython

from langchain.chat_models import init_chat_model
from lg_agents import SQLiteAgentPolicy, DockerConfig, SQLAgent

CHINOOK_DATABASE_HINTS="""
Here are some additional hints about tables and columns the database.

Table InvoiceLine:
    - Column UnitPrice is in US dollar units.

"""
class ChinookSQLitePolicy(SQLiteAgentPolicy):

    def __init__(self):
        db_file = 'Chinook_Sqlite.sqlite'
        if not os.path.isfile(db_file):
            print(f'Downloading {db_file} ...')            
            url = f'https://github.com/lerocha/chinook-database/releases/download/v1.4.5/{db_file}'
            urllib.request.urlretrieve(url, db_file)
        super().__init__(db_file)

    def database_hints(self):
        return CHINOOK_DATABASE_HINTS

def main():
    
    model = init_chat_model('openai:gpt-4.1')
    with DockerConfig() as dc, tempfile.TemporaryDirectory() as tempdir:
        policy = ChinookSQLitePolicy()
        agent = SQLAgent(
                agent_policy=policy, 
                docker_config=dc, 
                tmpdir=tempdir,
                show_artifacts=True,
                model=model
        )
        agent.chat('Which artists have the most revenue?')
        IPython.embed()

if __name__ == "__main__":
    main()



import os
import pdb
import tempfile
from typing import Literal, TypedDict, List, Optional
from uuid import uuid4
import abc
from dataclasses import dataclass

from langchain.chat_models import init_chat_model
from langchain_core.messages import AIMessage, SystemMessage, HumanMessage
from langchain_core.runnables import RunnableConfig
from langgraph.graph import END, START, MessagesState, StateGraph
from langgraph.checkpoint.memory import MemorySaver 
from langchain_core.runnables.graph_png import PngDrawer
from langchain_core.language_models import BaseChatModel
from pydantic import BaseModel, Field
from langchain.output_parsers import PydanticOutputParser
from langchain_core.tools import StructuredTool
from langgraph.types import Command
from langgraph.prebuilt import ToolNode
from tqdm import tqdm

from .python_toolkit import (
        PythonRunnerToolkit, DockerConfig, PythonRunnerToolContext,
        PythonRunnerTool
)
from .utils import show_image

class SQLAgentPolicy(abc.ABC):
    """
    This policy class allows an application to customize the SQLAgent for their
    use case. The SQLAgent allows an LLM to interact with a database that is
    exposed as a SQLAlchemy Engine object. The LLM code that interacts with the
    SQLAlchemy Engine object runs inside a Docker container.

    The Docker container has the following "work" folders:

    - `/home/pythonuser`: This is the working folder for any code executed by the LLM.
    - `/home/pythonuser/src`: Hosts any source files needed to expose the Engine object to
      the LLM. Typically, it contains a file (say, `database.py`) from which an
      LLM can import the engine object using a line such as `from src.database
      import engine`. Any additional files needed by the application (such as
      an sqlite database), can be copied to this folder by overriding the
      `files_to_copy` method.
    - `/home/pythonuser/artifacts`: This folder inside the container contains artifacts
      generated by the LLM code (such as PNG plots).

    At a minimum, a policy should override `engine_export_code` to define the
    source code that will allow an agent to import an engine object. For
    example, the following code will expose the `Chinook.db` sqlite data base
    as an SQLAlchemy Engine object called `engine` to any code that needs it.

    ```python
    import os
    from sqlalchemy import create_engine
    script_dir = os.path.dirname(os.path.abspath(__file__))
    db = os.path.join(script_dir, 'Chinook.db')
    engine = create_engine(f"sqlite:///{db}")
    __all__ = [ 'engine' ]
    ```

    The above, of course, assumes that the application has enabled the copying
    of `Chinook.db` file to the Docker container by overriding the
    `files_to_copy` method as follows:

    ```python
    def files_to_copy(tmpdir: str):
        return [ '/local/path/to/Chinook.db' ]

    Applications can also override `database_hints` method to return a string
    describing the database table entries (for example, what's the unit of the
    currency column?). If provided, the agent will provide this information to
    the LLM for its reference.
    """
    
    def files_to_copy(self) -> List[str]:
        """
        Specify a list of files to copy into the working directory of the
        Docker Container.
        """
        return []

    @abc.abstractmethod
    def engine_export_code(self) -> str:
        """
        Return the code that will expose an SQLAlchemy Engine object under the
        name `engine`. See the description of SQLAgentPolicy for more details.
        """

    def database_hints(self) -> Optional[str]:
        """
        A string that provides any hints that an application may want to
        provide the LLM about specific tables and columns. Don't include schema
        since that will be already provided to the LLM using the SQLAlchemy
        engine object. 

        For example, you can include some text to indicate that an "Invoice"
        column is in the units of dollars, or a "Time" column is in the units
        of "milliseconds".
        """
        return None

class SQLAgentState(MessagesState):
    first_time: bool

SQLA_LIST_TABLES_PYTHON=r"""
from src.database import engine
from sqlalchemy import MetaData

# Create a metadata object
metadata = MetaData()
metadata.reflect(bind=engine)

# Get the table names
table_names = metadata.tables.keys()
print('Available tables are:', list(table_names))
"""

SQLA_GET_SCHEMA_PYTHON=r"""
from src.database import engine
from sqlalchemy import MetaData
from sqlalchemy.schema import CreateTable

# Load metadata
metadata = MetaData()
metadata.reflect(bind=engine)

# Generate CREATE TABLE statements
for table_name, table in metadata.tables.items():
    create_statement = str(CreateTable(table).compile(engine))
    print(f"Schema for '{table_name} table':\n{create_statement}\n")

"""
def _run_python_tool(code: str, config:RunnableConfig):

    # create an artificial tool call and an AI message.
    tool_call = {
        "name": "run_python_tool",
        "args": {
            'python_code': code,
            'artifacts_abs_paths': [],
        },
        "id": uuid4().hex,
        "type": "tool_call",
    }
    tool_call_message = AIMessage(content="", tool_calls=[tool_call])

    # run the python tool on behalf of AI.
    config_idx: int = config['configurable'].get('config_idx', 0) # pyright: ignore[reportTypedDictNotRequiredAccess]
    python_runner = agent_globals_registry.get_globals(config_idx).agent.python_runner_tool
    tool_message = python_runner.invoke(tool_call, config=config)
    
    # return the request and response.
    return {"messages": [tool_call_message, tool_message]}

def list_tables(state: SQLAgentState, config: RunnableConfig):
    """
    Create a pre-determined tool call to list database table entries.
    """
    return _run_python_tool(SQLA_LIST_TABLES_PYTHON, config)

def call_get_schema(state: SQLAgentState, config: RunnableConfig):
    """
    Force the model to query the schema of the SQL database.
    """
    return _run_python_tool(SQLA_GET_SCHEMA_PYTHON, config)

class AnswerSchema(BaseModel):
    message: str = Field(description="""
    The answer to the user's question in markdown format. Include links to artifacts as appropriate.
    """)
    user_artifacts_abs_paths: List[str] = Field(description="""
    Local paths of artifacts (if any) produced as a part of the user's answer.
    """)

REACT_SYSTEM_PROMPT ="""
You are an agent designed to interact with a SQL database. You'll do so by
writing and executing python code using a python runner tool
("run_python_tool") that is equipped with Python 3.12+.

The database is exposed to you using an SQLAlchemy Engine object which you can
import as follows:

```
from src.database import engine
```

Always include the above line when referring to the databse. You can now use
the engine object imported above to list tables, schema etc., and also to run
queries.

The python tool returns the stdout of the program and hence you can use print
statements to get the information you want. 

Given an input question, create a syntactically correct code to run, then look
at the results of the run, and return the answer. You can also use the tool to
create artifiacts such as plots to show the user.

Here is an example of code that lists all the tables in the database:

```
{example_code}
```

You also have a available a `get_database_hints` tool that returns some hints
about the contents os the database.

DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.

{output_instructions}
"""

def react_node(state: SQLAgentState, config: RunnableConfig):
    """
    The main node in which we invoke the reasoning ability of the agent.
    """
    # get agent instance.
    config_idx: int = config['configurable'].get('config_idx', 0)  # pyright: ignore[reportTypedDictNotRequiredAccess]
    agent: SQLAgent = agent_globals_registry.get_globals(config_idx).agent
    # create a system prompt.
    system_message = SystemMessage(agent.system_prompt)
    # bind the python runner tool the the LLM.
    llm_with_tools = agent.llm.bind_tools([agent.python_runner_tool, get_database_hints])
    # invoke the LLM.
    response = llm_with_tools.invoke([system_message] + state["messages"])
    return {"messages": [response]}

def route_tool(state: SQLAgentState) -> Literal[END, 'run_python']:
    messages = state["messages"]
    last_message = messages[-1]
    assert isinstance(last_message, AIMessage)
    tool_calls = last_message.tool_calls
    if len(tool_calls) == 0:
        return END
    else:
        return 'run_python'

def should_get_table_meta(state: SQLAgentState, config: RunnableConfig) -> Command[Literal["react_node", "list_tables"]]:
    
    # if this is not the first time, then go directly to the react node and
    # don't update any state.
    if 'first_time' in state and state['first_time'] == 0:
        return Command(update={}, goto='react_node')
    # if this is the first time, go to list_tables and clear first_time.
    # - also, add runtime configuration to the state, so it available to all
    #   the tools.
    else:
        return Command(
                update = {
                    'first_time': 0,
                    }, 
                goto="list_tables"
        )


SQLITE_DATABASE_LOADER_CODE="""
import os
from sqlalchemy import create_engine
script_dir = os.path.dirname(os.path.abspath(__file__))
db = os.path.join(script_dir, '{db_filename}')
engine = create_engine("sqlite:///" + db)
__all__ = [ 'engine' ]
"""

class SQLiteAgentPolicy(SQLAgentPolicy):
    """
    A helper for SQLite databases that includes some default implementation
    code that pertains to SQLite databases.
    """

    def __init__(self, db_file: str):
        self.db_file = db_file

    def files_to_copy(self):
        return [ self.db_file ]

    def engine_export_code(self):
        return SQLITE_DATABASE_LOADER_CODE.format(
                db_filename = os.path.basename(self.db_file)
        )

def _get_database_hints(config: RunnableConfig) -> str:
    config_idx: int = config['configurable'].get('config_idx', 0)  # pyright: ignore[reportTypedDictNotRequiredAccess]
    agent_policy: SQLAgentPolicy = agent_globals_registry.get_globals(config_idx).agent.agent_policy
    hints = agent_policy.database_hints() 
    if hints is not None:
        return hints
    else:
        return "No hints are available."

GET_DATABASE_HINTS_DESC="""
Returns some hints about tables and columns in the database.

Returns:
    str: Hints about the tables and columns in a database.
"""
get_database_hints = StructuredTool.from_function(
        func=_get_database_hints,
        name="get_database_tool",
        description=GET_DATABASE_HINTS_DESC,
        return_direct=True
)

def database_hints_node(state: SQLAgentState, config: RunnableConfig):
    """
    Create a pre-determined tool call to provide hints about the database.
    """
    tool_call = {
        "name": "get_database_hints",
        "args": {},
        "id": uuid4().hex,
        "type": "tool_call",
    }
    tool_call_message = AIMessage(content="", tool_calls=[tool_call])
    tool_message = get_database_hints.invoke(tool_call, config=config)
    return {"messages": [tool_call_message, tool_message]}

class SQLAgentConfigSchema(TypedDict):
    config_idx: int  

class SQLAgent:

    def __init__(self, 
                 agent_policy:SQLAgentPolicy,
                 docker_config: DockerConfig,
                 tmpdir: str,
                 model: str | BaseChatModel = "openai:gpt-4.1",
                 show_artifacts: bool =False,
                 recursion_limit: int =25
    ):
        """
        Creates a new SQL Agent.

        Args:
            agent_policy: App specific customization for the agent.
            docker_config: Docker configuration for Python Tool.
            tmpdir: A temporary directory to use for the temporary files
                created during the agent run (recommend using
                tempfile.TemporaryDirectory() context manager for this).
            model: The model to use.
            recursion_limit: The max number of graph state transitions to
                execute.
        """

        # latch parameters.
        self.recursion_limit = recursion_limit

        # initialize LLM.
        if isinstance(model, str):
            self.llm = init_chat_model(model)
        else:
            self.llm = model

        # configure the python runner.
        self.python_tool = PythonRunnerToolContext(
                docker_config,
                tmpdir,
                ignore_dependencies=['src'],
                ignore_unsafe_functions=['compile']
        )
        self.python_toolkit = PythonRunnerToolkit.from_context(
                config=self.python_tool
        )
        # python toolkit only exposes a single tool
        self.python_runner_tool: PythonRunnerTool = next(
                tool 
                for tool in self.python_toolkit.get_tools() 
                if tool.name == "run_python_tool"
        ) # pyright: ignore[reportAttributeAccessIssue]

        # latch agent policy.
        self.agent_policy: SQLAgentPolicy = agent_policy

        # initialize the database configuration.
        # - copy any files the application needs.
        for file in self.agent_policy.files_to_copy():
            self.python_tool.copy_file_to_container(file, '/home/pythonuser/src')
        # -- copy the database loader code to container.
        self.python_tool.copy_code_to_container(
                self.agent_policy.engine_export_code(),
                '/home/pythonuser/src/database.py'
        )

        # graph building.
        self._build_graph()

        # initialize runtime configuration.
        self._init_config()

        # generate system prompt for react agent.
        self.show_artifacts = show_artifacts
        if show_artifacts:
            output_instructions = PydanticOutputParser(pydantic_object=AnswerSchema).get_format_instructions()
        else:
            output_instructions = "Return your answer with Markdown formatting."
        self.system_prompt = REACT_SYSTEM_PROMPT.format(
            example_code=SQLA_LIST_TABLES_PYTHON,
            output_instructions=output_instructions
        )

    def _set_globals(self) -> int:
        
        return agent_globals_registry.register_globals(self)

    def _build_graph(self):

        self.config_idx = self._set_globals()
        builder = StateGraph(SQLAgentState, config_schema=SQLAgentConfigSchema)

        # are hints available?
        hints_avail = self.agent_policy.database_hints() is not None

        # add nodes.
        builder.add_node(should_get_table_meta)
        builder.add_node(list_tables)
        builder.add_node(call_get_schema)
        builder.add_node(react_node)
        builder.add_node(ToolNode([self.python_runner_tool], name="run_python"))
        if hints_avail:
            builder.add_node(database_hints_node)

        # connect them.
        builder.add_edge(START, "should_get_table_meta")
        builder.add_edge("list_tables", "call_get_schema")
        if hints_avail:
            builder.add_edge("call_get_schema", "database_hints_node")
            builder.add_edge("database_hints_node", "react_node")
        else:
            builder.add_edge("call_get_schema", "react_node")
        builder.add_conditional_edges(
            "react_node",
            route_tool,
        )
        builder.add_edge("run_python", "react_node")

        self.memory = MemorySaver()
        self.agent = builder.compile(checkpointer=self.memory)

    def print_ascii(self):
        self.agent.get_graph().print_ascii()

    def display_graph(self):

        with tempfile.TemporaryDirectory() as tmpdir:
            drawer = PngDrawer()
            png_path = os.path.join(tmpdir, '_lang_graph.png')
            drawer.draw(self.agent.get_graph(), png_path)
            show_image(png_path)

    def display_artifacts(self, artifacts: List[str]):
        """
        Displays artifacts generated by the Model.
        """
        for artifact in artifacts:
            if not os.path.isfile(artifact):
                raise RuntimeError(f'Artifact {artifact} is not found!')
            _, ext = os.path.splitext(artifact)
            match ext:
                case '.png': show_image(artifact)

    def artifacts(self) -> List[str]:
        """
        Returns the list of artifacts from the last AI message.
        """
        final_state = self.agent.get_state(self.config) # pyright: ignore[reportArgumentType]
        final_message = final_state.values['messages'][-1]
        answer: AnswerSchema = AnswerSchema.model_validate_json(final_message.content)
        return answer.user_artifacts_abs_paths

    def _init_config(self):

        self.config = {
                'recursion_limit': self.recursion_limit,
                'configurable': {
                    'thread_id': 1,
                    'config_idx': self.config_idx
                }
        }

    def chat(self, question:str, quiet=True):

        gen = self.agent.stream(
            {"messages": [HumanMessage(question)]}, 
            stream_mode="values", 
            config=self.config # type: ignore
        )
        gen = tqdm(gen, desc="Thinking ...", total=self.recursion_limit) if quiet else gen
        for step in gen:
            if not quiet:
                step["messages"][-1].pretty_print()

        print('================================== Final Answer ================================== ')
        final_state = self.agent.get_state(self.config) # pyright: ignore[reportArgumentType]
        final_message = final_state.values['messages'][-1]
        if self.show_artifacts:
            answer: AnswerSchema = AnswerSchema.model_validate_json(final_message.content)
            print(answer.message)
            self.display_artifacts(answer.user_artifacts_abs_paths)
        else:
            print(final_message.content)

@dataclass
class SQLAgentGlobals:
    """
    A global variable for tools to access non-trivial agent related objects.
    The config object will just contain an index of the global object
    associated with the agent.
    """
    agent: SQLAgent

class SQLAgentGlobalsRegistry:
    """
    A registry to keep track of every SQLAgentGlobal object created.
    """
    def __init__(self):
        self.globals: List[SQLAgentGlobals] = []

    def register_globals(
            self,
            agent: SQLAgent
            ):
        idx = len(self.globals)
        self.globals.append(SQLAgentGlobals(agent))
        return idx

    def get_globals(self, index) -> SQLAgentGlobals:
        return self.globals[index]

# a registry to keep track of non-trivial objects needed by the nodes and
# tools.
agent_globals_registry = SQLAgentGlobalsRegistry()

